#include <cmath>

#include "falconmodel.h"
#include "modelerdraw.h"
#include "armmodel.h"

#define _USE_MATH_DEFINES // for M_PI

ArmModel::ArmModel(const char* name) :
	Model(name),
	arm_rotation_x("Arm Rotation X", -180, 180, 0, 1),
	arm_rotation_y("Arm Rotation Y", -90, 90, 0, 1),
	upper_arm_angle("Upper Arm Angle", -90, 90, 0, 1),
	lower_arm_angle("Lower Arm Angle", 0, 180, 0, 1),
	hand_angle("Hand Angle", -90, 90, 0, 1),
	upper_arm_color("Upper Arm Color", 0, 0, 1),
	lower_arm_color("Lower Arm Color", 1, 1, 0),
	hand_color("Hand Color", 1, 1, 0)
{
	properties.add(&arm_rotation_x);
	properties.add(&arm_rotation_y);
	properties.add(&upper_arm_angle);
	properties.add(&lower_arm_angle);
	properties.add(&hand_angle);
	properties.add(&upper_arm_color);
	properties.add(&lower_arm_color);
	properties.add(&hand_color);
	
	muscle_mass = 1.0;
}

void ArmModel::draw()
{
	// upper arm
	setDiffuseColor(upper_arm_color.getRed(),
		upper_arm_color.getGreen(),
		upper_arm_color.getBlue());
	glPushMatrix();
	float rotation_x = arm_rotation_x.getValue();
	glRotatef(rotation_x, 0, 0, 1);
	float ua_angle = upper_arm_angle.getValue();
	glRotatef(ua_angle, 1, 0, 0);
	// scale for muscle
	glPushMatrix();
	glScaled(muscle_mass, muscle_mass, 1);
	drawCylinder(UPPER_ARM_LENGTH, UPPER_ARM_RADIUS, ELBOW_RADIUS);
	glPopMatrix();

	// lower arm
	setDiffuseColor(lower_arm_color.getRed(),
		lower_arm_color.getGreen(),
		lower_arm_color.getBlue());
	glPushMatrix();
	float la_angle = lower_arm_angle.getValue();
	glTranslated(0, 0, UPPER_ARM_LENGTH);
	glRotatef(-1 * la_angle, 1, 0, 0);
	glPushMatrix();
	glScaled(muscle_mass, muscle_mass, 1);
	drawCylinder(LOWER_ARM_LENGTH, ELBOW_RADIUS, LOWER_ARM_RADIUS);
	glPopMatrix();

	// hand
	setDiffuseColor(hand_color.getRed(),
		hand_color.getGreen(),
		hand_color.getBlue());
	glPushMatrix();
	float h_angle = hand_angle.getValue();
	glTranslated(0, 0, LOWER_ARM_LENGTH);
	glRotatef(h_angle, 1, 0, 0);
	glTranslated(0, HAND_DEPTH / -2, HAND_LENGTH / 2);
	drawPrism(HAND_WIDTH, HAND_WIDTH, HAND_DEPTH, HAND_LENGTH, HAND_LENGTH);
	glPopMatrix();

	glPopMatrix();

	glPopMatrix();
}