// a model of captain falcon
#include <cmath>

#include "modelerdraw.h"
#include "armmodel.h"
#include "legmodel.h"
#include "headmodel.h"
#include "falconmodel.h"

#define _USE_MATH_DEFINES // for M_PI

FalconModel::FalconModel(const char* name) :
	Model(name),
	torso_rotation("Torso Rotation", 0, 360, 0, 1),
	muscle_mass("Muscle Mass", 0.25f, 2.f, 1.f, 0.05f),
	right_arm("Right Arm"),
	left_arm("Left Arm"),
	right_leg("Right Leg"),
	left_leg("Left Leg"),
	head("Head"),
	color("Torso Color", 0, 0, 1)
{
	properties.add(&torso_rotation);
	properties.add(&muscle_mass);
	properties.add(&color);
	properties.add(right_arm.getProperties());
	properties.add(left_arm.getProperties());
	properties.add(right_leg.getProperties());
	properties.add(left_leg.getProperties());
	properties.add(head.getProperties());

	torso_height = TORSO_HEIGHT;
}

void FalconModel::draw() {
	float mm = muscle_mass.getValue();
	right_arm.muscle_mass = mm;
	left_arm.muscle_mass = mm;
	left_leg.muscle_mass = mm;
	right_leg.muscle_mass = mm;

	// update these values to scale for muscle mass.
	// can't just use a glScaled because it affects how we
	// translate all the other pieces, and we don't want
	// to scale them the same way.
	torso_base_width = TORSO_BASE_WIDTH * mm;
	torso_top_width = TORSO_TOP_WIDTH * mm;
	torso_base_depth = TORSO_BASE_DEPTH * mm;
	torso_top_depth = TORSO_TOP_DEPTH * mm;

	// its big af rn so scale it all down a lil
	glScaled(0.25, 0.25, 0.25);

	// torso
	setDiffuseColor(color.getRed(),
		color.getGreen(),
		color.getBlue());
	glPushMatrix();
	glTranslated(0, UPPER_LEG_LENGTH + LOWER_LEG_LENGTH, 0);
	glRotatef(torso_rotation.getValue(), 0, 1, 0);
	drawPrism(torso_base_width, torso_top_width, torso_height,
						 torso_base_depth, torso_top_depth);

	// right arm
	glPushMatrix();
	glTranslated(torso_top_width / 2, torso_height - UPPER_ARM_RADIUS, 0);
	glRotatef(-1 * right_arm.arm_rotation_y.getValue(), 0, 1, 0);
	// the arm is drawn along the positive z axis - we want to rotate it
	// to point in the positive x direction.
	glRotatef(90, 0, 1, 0);
	right_arm.draw();
	glPopMatrix();

	// left arm
	glPushMatrix();
	glTranslated(torso_top_width / -2, torso_height - UPPER_ARM_RADIUS, 0);
	glRotatef(left_arm.arm_rotation_y.getValue(), 0, 1, 0);
	// this time rotate in the opposite direction
	glRotatef(-90, 0, 1, 0);
	left_arm.draw();
	glPopMatrix();

	// right leg
	glPushMatrix();
	glTranslated(torso_base_width / 2 - UPPER_LEG_RADIUS / 1.5, 0, 0);
	// leg is drawn in positive z direction, so we want to rotate it 90 degrees
	// counter clockwise about the positive x axis to point in the negative
	// y direction
	glRotatef(90, 1, 0, 0);
	right_leg.draw();
	glPopMatrix();

	// left leg
	glPushMatrix();
	glTranslated(torso_base_width / -2 + UPPER_LEG_RADIUS / 1.5, 0, 0);
	glRotatef(90, 1, 0, 0);
	left_leg.draw();
	glPopMatrix();

	// head
	glPushMatrix();
	glTranslated(0, torso_height + HEAD_RADIUS, 0);
	head.draw();
	glPopMatrix();

	glPopMatrix();
}