#include <cmath>

#include "modelerdraw.h"
#include "headmodel.h"

#define _USE_MATH_DEFINES // for M_PI

HeadModel::HeadModel(const char* name) :
	Model(name),
	head_rotation("Head Rotation", -90, 90, 0, 1),
	head_angle("Head Angle Angle", -30, 30, 0, 1),
	head_color("Head Color", 1, 0, 0),
	headpiece_color("Headpiece Color", 1, 1, 0)
{
	properties.add(&head_rotation);
	properties.add(&head_angle);
	properties.add(&head_color);
	properties.add(&headpiece_color);
}

void HeadModel::draw()
{
	// draw head
	setDiffuseColor(head_color.getRed(),
		head_color.getGreen(),
		head_color.getBlue());
	glPushMatrix();
	glRotatef(head_rotation.getValue(), 0, 1, 0);
	glRotatef(head_angle.getValue(), 1, 0, 0);
	glScaled(1, 1.10, 1);  // stretch y for an elliptical head
	glutSolidSphere(HEAD_RADIUS, SPHERE_DIVISIONS, SPHERE_DIVISIONS);

	// draw headpiece - basically two trapezoids stuck together at 45 degree angles
	// left (i think?) part first
	setDiffuseColor(headpiece_color.getRed(),
		headpiece_color.getGreen(),
		headpiece_color.getBlue());
	glPushMatrix();
	glTranslated(0, HEAD_RADIUS / 2, HEAD_RADIUS);
	glRotatef(-45, 0, 0, 1);
	glTranslated(HEADPIECE_BASE_WIDTH / -2, 0, 0);
	drawPrism(HEADPIECE_BASE_WIDTH, HEADPIECE_TOP_WIDTH, HEADPIECE_HEIGHT,
			  HEADPIECE_DEPTH, HEADPIECE_DEPTH);
	glPopMatrix();

	// right part
	glPushMatrix();
	glTranslated(0, HEAD_RADIUS / 2, HEAD_RADIUS);
	glRotatef(45, 0, 0, 1);
	glTranslated(HEADPIECE_BASE_WIDTH / 2, 0, 0);
	drawPrism(HEADPIECE_BASE_WIDTH, HEADPIECE_TOP_WIDTH, HEADPIECE_HEIGHT,
			  HEADPIECE_DEPTH, HEADPIECE_DEPTH);
	glPopMatrix();

	glPopMatrix();
}