#include <cmath>

#include "modelerdraw.h"
#include "legmodel.h"

#define _USE_MATH_DEFINES // for M_PI

LegModel::LegModel(const char* name) :
	Model(name),
	leg_rotation("Leg Rotation", -30, 30, 0, 1),
	upper_leg_angle("Upper Leg Angle", -90, 90, 0, 1),
	lower_leg_angle("Lower Leg Angle", 0, 180, 0, 1),
	foot_angle("Foot Angle", 0, 120, 90, 1),
	upper_leg_color("Upper Leg Color", 0, 0, 1),
	lower_leg_color("Lower Leg Color", 1, 1, 0),
	foot_color("Foot Color", 1, 1, 0)
{
	properties.add(&leg_rotation);
	properties.add(&upper_leg_angle);
	properties.add(&lower_leg_angle);
	properties.add(&foot_angle);
	properties.add(&upper_leg_color);
	properties.add(&lower_leg_color);
	properties.add(&foot_color);
}

void LegModel::draw()
{
	// draw upper leg
	setDiffuseColor(upper_leg_color.getRed(),
		upper_leg_color.getGreen(),
		upper_leg_color.getBlue());
	glPushMatrix();
	float rotation = leg_rotation.getValue();
	glRotatef(rotation, 0, 0, 1);
	float ul_angle = upper_leg_angle.getValue();
	glRotatef(ul_angle, 1, 0, 0);
	// scale for muscle
	glPushMatrix();
	glScaled(muscle_mass, muscle_mass, 1);
	drawCylinder(UPPER_LEG_LENGTH, UPPER_LEG_RADIUS, KNEE_RADIUS);
	glPopMatrix();

	// draw lower leg
	setDiffuseColor(lower_leg_color.getRed(),
		lower_leg_color.getGreen(),
		lower_leg_color.getBlue());
	glPushMatrix();
	float ll_angle = lower_leg_angle.getValue();
	glTranslated(0, 0, UPPER_LEG_LENGTH);
	glRotatef(ll_angle, 1, 0, 0);
	// scale for muscle
	glPushMatrix();
	glScaled(muscle_mass, muscle_mass, 1);
	drawCylinder(LOWER_LEG_LENGTH, KNEE_RADIUS, LOWER_LEG_RADIUS);
	glPopMatrix();

	// draw foot
	setDiffuseColor(foot_color.getRed(),
		foot_color.getGreen(),
		foot_color.getBlue());
	glPushMatrix();
	float f_angle = foot_angle.getValue();
	glTranslated(0, 0, LOWER_LEG_LENGTH);
	glRotatef(-1 * f_angle, 1, 0, 0);
	glTranslated(0, FOOT_HEIGHT / -2, FOOT_LENGTH / 2 - LOWER_LEG_RADIUS);
	drawPrism(FOOT_WIDTH, FOOT_WIDTH, FOOT_HEIGHT, FOOT_LENGTH, FOOT_LENGTH);
	glPopMatrix();

	glPopMatrix();

	glPopMatrix();
}